#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>           /* struct timeval */

#include "tableSTAT.h"

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

extern int errno;
void err();
void mes();
int Nread();
int Nwrite();

void sigpipe() {  signal(SIGPIPE, sigpipe); }

void sigchld()
{
 int sst;

 wait( &sst );
 signal( SIGCHLD, sigchld );
}

int fd= -1;                         /* fd of network socket */
int fd_cmd= -1;

void sigterm()
{
 mes( "*** EXIT: got SIGTERM! ***" );

 if( fd_cmd != -1 )  close( fd_cmd );
 if( fd != -1 )  close( fd );

 exit(0);
}

Usage( char *name )  { fprintf( stderr, "Usage: %s -k shmkey [-p port (5001 default)]\n", name ); exit( -2 ); }



main( int argc, char *argv[] )
{
 int domain, fromlen;
 short port= 5001;               /* TCP port number */
 int i, cnt, key= -1;
 int shmid, statID;
 tableSTAT st, *stp;


 struct sockaddr_in peer;
 int peerlen= sizeof(peer);

 for( i= 1; i < argc; i++ )
 {
  if( strcmp( argv[i], "-p" ) == 0 )
  {
   port= atoi( argv[++i] );
   continue;
  }

  if( strcmp( argv[i], "-k" ) == 0 )
  {
   key= atoi( argv[++i] );
   continue;
  }

  Usage( argv[0] );
 }

 if( key == -1 )  Usage( argv[0] );

 sinme.sin_port=  htons(port);
#if defined(__convex__)
 sinme.sin_family= AF_INET;
 sinme.sin_addr.s_addr= INADDR_ANY ;
#endif

 while(1)
 {
  if((fd= socket(AF_INET, SOCK_STREAM, 0)) < 0 )    err("socket");
  if( bind( fd, &sinme, sizeof(sinme) ) >= 0 )  break;
  mes( "-- Bind Error, will try again..." );
  sleep( 30 );
 }

 signal(SIGPIPE, sigpipe);
 signal(SIGCHLD, sigchld);
 signal(SIGTERM, sigterm);

 listen(fd,0);   /* allow a queue of 0 */

 if( (shmid= shmget( key, sizeof( tableSTAT ), IPC_CREAT | 0777 )) == -1 )
 {
  puts( "### Cannot get SHM..." );
  close( fd );
  exit( -2 );
 }

 if( (stp= (tableSTAT *)shmat( shmid, 0, 0 )) == (tableSTAT *)-1 )
 {
  puts( "### Cannot attach SHM..." );
  close( fd );
  exit( -2 );
 }

 memset( stp, 0, sizeof( tableSTAT ));
 stp->number= 0;

 mes( "*** READY! ***" );

 fromlen= sizeof(frominet);
 domain= AF_INET;

 while( 1 )
 {
  int cnt;
  char name[100];

  if((fd_cmd= accept( fd, &frominet, &fromlen )) < 0 )
  {
   if( errno == EINTR )  continue;
   err("accept");
  }

  if( (cnt= Nread( fd_cmd, name, sizeof(name) )) <= 0 )
  {
   mes( "Nread" );
   close( fd_cmd );
   continue;
  }

  name[cnt]= '\0';
  for( statID= 0; statID < TABSIZE; statID++ )
   if( strcmp( stp->stat[ statID ].name, name ) == 0 )
    break;

  if( statID >= TABSIZE )
  {
   if( stp->number >= TABSIZE-1 )
   {
    fprintf( stderr, "### Sorry, cannot listen more than: %d STAT(s)\n", TABSIZE );
    close( fd_cmd );
    continue;
   }

   statID= stp->number++;
  }

  strcpy( stp->stat[ statID ].name, name );
  stp->stat[ statID ].active= 0;

  if( fork() == 0 )
  {
   if( getpeername( fd_cmd, (struct sockaddr_in *) &peer, &peerlen ) < 0 )  err("getpeername");
   anySTAT( fd_cmd,  inet_ntoa(peer.sin_addr), &(stp->stat[ statID ]) );
   stp->stat[ statID ].active= 0;
   close( fd_cmd );

   exit(0);
  }

  close( fd_cmd );
 }
}


anySTAT( int fd, char *addr, counterSTAT *stat )
{
 long T;
 int cnt;

 errno= 0;
 time( &T );
 fprintf( stderr, "anySTAT-%d:> Connect -> [%s] STAT= \"%s\" TIME: %s", getpid(), addr, stat->name, ctime(&T) );
 fflush( stderr );

 while( 1 )
 {
  if( (cnt= Nread( fd, stat->values, sizeof( stat->values) )) <= 0 )  break;
  stat->values[ cnt ]= 0;
  stat->active= 1;
 }

 close( fd );
 stat->active= 0;

 time( &T );
 fprintf( stderr, "anySTAT-%d:> Exit -> [%s] STAT= \"%s\" TIME: %s", getpid(), addr, stat->name, ctime(&T) );
 fflush( stderr );
}



void err( char *s)
{
 perror(s);
 fprintf( stderr," errno= %d\n", errno );
 fflush( stderr );
 exit( -2 );
}

void mes( char *s )
{
 fprintf( stderr,"anySTAT-%d: %s\n", getpid(), s );
 fflush( stderr );
}



/*
 *                      N R E A D
 */
Nread( fd, buf, count )
int fd;
char *buf;
int count;
{
 int cnt;
 cnt= read( fd, buf, count );
 if( cnt >= 0 ) buf[cnt]= 0;
 return( cnt );
}

/*
 *                      N W R I T E
 */
Nwrite( fd, buf, count )
int fd;
char *buf;
int count;
{
 return( write( fd, buf, count ) );
}

